import subprocess
from datetime import datetime
from multiprocessing import Manager, Queue
from os import PathLike
from pathlib import Path
from queue import Empty

from typing import Dict, List, Optional

from skyplane.chunk import ChunkRequest, ChunkState
from skyplane.utils import logger


class ChunkStore:
    def __init__(self, chunk_dir: PathLike):
        self.chunk_dir = Path(chunk_dir)
        self.chunk_dir.mkdir(parents=True, exist_ok=True)

        # delete existing chunks
        for chunk_file in self.chunk_dir.glob("*.chunk"):
            logger.warning(f"Deleting existing chunk file {chunk_file}")
            chunk_file.unlink()

        # multiprocess-safe concurrent structures
        self.manager = Manager()
        self.chunk_requests: Dict[str, ChunkRequest] = self.manager.dict()  # type: ignore
        self.chunk_status: Dict[str, ChunkState] = self.manager.dict()  # type: ignore

        # state log
        self.chunk_status_queue: Queue[Dict] = Queue()

        # metric log
        self.sender_compressed_sizes: Dict[str, float] = self.manager.dict()  # type: ignore

    def get_chunk_file_path(self, chunk_id: str) -> Path:
        return self.chunk_dir / f"{chunk_id}.chunk"

    ###
    # ChunkState management
    ###
    def get_chunk_state(self, chunk_id: str) -> Optional[ChunkState]:
        return self.chunk_status[chunk_id] if chunk_id in self.chunk_status else None

    def set_chunk_state(self, chunk_id: str, new_status: ChunkState, log_metadata: Optional[Dict] = None):
        self.chunk_status[chunk_id] = new_status
        rec = {"chunk_id": chunk_id, "state": new_status.name, "time": str(datetime.utcnow().isoformat())}
        if log_metadata is not None:
            rec.update(log_metadata)
        self.chunk_status_queue.put(rec)

    def drain_chunk_status_queue(self) -> List[Dict]:
        out_events = []
        while True:
            try:
                elem = self.chunk_status_queue.get_nowait()
                out_events.append(elem)
            except Empty:
                break
        return out_events

    def state_queue_download(self, chunk_id: str):
        state = self.get_chunk_state(chunk_id)
        if state in [ChunkState.registered, ChunkState.download_queued]:
            self.set_chunk_state(chunk_id, ChunkState.download_queued)
        else:
            raise ValueError(f"Invalid transition queue_download from {state} (id={chunk_id})")

    def state_start_download(self, chunk_id: str, receiver_id: Optional[str] = None):
        state = self.get_chunk_state(chunk_id)
        if state in [ChunkState.download_queued, ChunkState.download_in_progress]:
            self.set_chunk_state(chunk_id, ChunkState.download_in_progress, {"receiver_id": receiver_id})
        else:
            raise ValueError(f"Invalid transition start_download from {state}")

    def state_finish_download(self, chunk_id: str, receiver_id: Optional[str] = None):
        state = self.get_chunk_state(chunk_id)
        if state in [ChunkState.download_in_progress, ChunkState.downloaded]:
            self.set_chunk_state(chunk_id, ChunkState.downloaded, {"receiver_id": receiver_id})
        else:
            raise ValueError(f"Invalid transition finish_download from {state} (id={chunk_id})")

    def state_queue_upload(self, chunk_id: str):
        state = self.get_chunk_state(chunk_id)
        if state in [ChunkState.downloaded, ChunkState.upload_queued]:
            self.set_chunk_state(chunk_id, ChunkState.upload_queued)
        else:
            raise ValueError(f"Invalid transition upload_queued from {state} (id={chunk_id})")

    def state_start_upload(self, chunk_id: str, sender_id: Optional[str] = None):
        state = self.get_chunk_state(chunk_id)
        if state in [ChunkState.upload_queued, ChunkState.upload_in_progress]:
            self.set_chunk_state(chunk_id, ChunkState.upload_in_progress, {"sender_id": sender_id})
        else:
            raise ValueError(f"Invalid transition start_upload from {state} (id={chunk_id})")

    def state_finish_upload(self, chunk_id: str, sender_id: Optional[str] = None, compressed_size_bytes: Optional[int] = None):
        state = self.get_chunk_state(chunk_id)
        if state in [ChunkState.upload_in_progress, ChunkState.upload_complete]:
            self.set_chunk_state(chunk_id, ChunkState.upload_complete, {"sender_id": sender_id})
            if compressed_size_bytes is not None:
                self.sender_compressed_sizes[chunk_id] = compressed_size_bytes
        else:
            raise ValueError(f"Invalid transition finish_upload from {state} (id={chunk_id})")

    def state_fail(self, chunk_id: str):
        if self.get_chunk_state(chunk_id) != ChunkState.upload_complete:
            self.set_chunk_state(chunk_id, ChunkState.failed)
        else:
            raise ValueError(f"Invalid transition fail from {self.get_chunk_state(chunk_id)} (id={chunk_id})")

    ###
    # Chunk management
    ###
    def get_chunk_requests(self, status: Optional[ChunkState] = None) -> List[ChunkRequest]:
        if status is None:
            return list(self.chunk_requests.values())
        else:
            return [req for i, req in self.chunk_requests.items() if self.get_chunk_state(i) == status]

    def get_chunk_request(self, chunk_id: str) -> ChunkRequest:
        if chunk_id not in self.chunk_requests:
            raise ValueError(f"ChunkRequest {chunk_id} not found")
        return self.chunk_requests[chunk_id]

    def add_chunk_request(self, chunk_request: ChunkRequest, state=ChunkState.registered):
        self.set_chunk_state(chunk_request.chunk.chunk_id, state)
        self.chunk_requests[chunk_request.chunk.chunk_id] = chunk_request

    def update_chunk_checksum(self, chunk_id: str, checksum: Optional[bytes]):
        cr = self.chunk_requests[chunk_id]
        cr.chunk.checksum = checksum
        self.chunk_requests[chunk_id] = cr

    def update_chunk_mime_type(self, chunk_id: str, mime_type: Optional[str]):
        cr = self.chunk_requests[chunk_id]
        cr.chunk.mime_type = mime_type
        self.chunk_requests[chunk_id] = cr

    # Memory space calculation
    def remaining_bytes(self):
        return int(subprocess.check_output(["df", "-k", "--output=avail", self.chunk_dir]).decode().strip().split()[-1]) * 1024



from os import chmod
from pathlib import Path

from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from skyplane.utils.fn import PathLike


def generate_keypair(pubkey_path: PathLike, pem_path: PathLike):
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=4096,
    )
    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM, crypto_serialization.PrivateFormat.TraditionalOpenSSL, crypto_serialization.NoEncryption()
    )
    public_key = key.public_key().public_bytes(crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH)
    Path(pubkey_path).write_bytes(public_key)
    Path(pem_path).write_bytes(private_key)
    chmod(pem_path, 0o600)



import warnings
from functools import lru_cache
from pathlib import Path

from cryptography.utils import CryptographyDeprecationWarning

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)
    import paramiko

from skyplane import exceptions
from skyplane.compute.gcp.gcp_auth import GCPAuthentication
from skyplane.compute.server import Server, ServerState, key_root
from skyplane.utils.fn import PathLike


class GCPServer(Server):
    def __init__(self, region_tag: str, instance_name: str, key_root: PathLike = key_root / "gcp", log_dir=None, ssh_private_key=None):
        super().__init__(region_tag, log_dir=log_dir)
        assert self.region_tag.split(":")[0] == "gcp", f"Region name doesn't match pattern gcp:<region> {self.region_tag}"
        self.gcp_region = self.region_tag.split(":")[1]
        self.auth = GCPAuthentication()
        self.gcp_instance_name = instance_name
        key_root = Path(key_root)
        key_root.mkdir(parents=True, exist_ok=True)
        if ssh_private_key is None:
            self.ssh_private_key = key_root / f"skyplane-gcp-cert.pem"
        else:
            self.ssh_private_key = ssh_private_key

    def uuid(self):
        return f"{self.region_tag}:{self.gcp_instance_name}"

    @lru_cache(maxsize=1)
    def get_gcp_instance(self):
        instances = self.auth.get_gcp_instances(self.gcp_region)
        if "items" in instances:
            for i in instances["items"]:
                if i["name"] == self.gcp_instance_name:
                    return i
        raise ValueError(f"No instance found with name {self.gcp_instance_name}, {instances}")

    def get_instance_property(self, prop):
        instance = self.get_gcp_instance()
        if prop in instance:
            return instance[prop]
        else:
            return None

    def public_ip(self):
        """Get public IP for instance with GCP client"""
        return self.get_instance_property("networkInterfaces")[0]["accessConfigs"][0].get("natIP")

    def private_ip(self):
        return self.get_instance_property("networkInterfaces")[0]["networkIP"]

    def instance_class(self):
        machine_type = self.get_instance_property("machineType")
        return machine_type.split("/")[-1] if machine_type else None

    def region(self):
        return self.gcp_region

    def instance_state(self):
        return ServerState.from_gcp_state(self.get_instance_property("status"))

    def instance_name(self):
        return self.get_instance_property("name")

    def tags(self):
        """Get labels for instance."""
        return self.get_instance_property("labels") or {}

    def network_tier(self):
        interface = self.get_instance_property("networkInterfaces")[0]
        return interface["accessConfigs"][0]["networkTier"]

    def __repr__(self):
        return f"GCPServer(region_tag={self.region_tag}, instance_name={self.gcp_instance_name})"

    def terminate_instance_impl(self):
        self.auth.get_gcp_client().instances().delete(
            project=self.auth.project_id, zone=self.gcp_region, instance=self.instance_name()
        ).execute()

    def get_ssh_client_impl(self, uname="skyplane", ssh_key_password="skyplane"):
        """Return paramiko client that connects to this instance."""
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            ssh_client.connect(
                hostname=self.public_ip(),
                username=uname,
                pkey=paramiko.RSAKey.from_private_key_file(str(self.ssh_private_key), password=ssh_key_password),
                look_for_keys=False,
                banner_timeout=200,
            )
            return ssh_client
        except paramiko.AuthenticationException as e:
            raise exceptions.BadConfigException(
                f"Failed to connect to GCP server {self.uuid()}. Delete local GCP keys and retry: `rm -rf {key_root / 'gcp'}`"
            ) from e

    def get_sftp_client(self, uname="skyplane", ssh_key_password="skyplane"):
        t = paramiko.Transport((self.public_ip(), 22))
        pkey = paramiko.RSAKey.from_private_key_file(str(self.ssh_private_key), password=ssh_key_password)
        t.connect(username=uname, pkey=pkey)
        return paramiko.SFTPClient.from_transport(t)

    def open_ssh_tunnel_impl(self, remote_port, uname="skyplane", ssh_key_password="skyplane"):
        import sshtunnel

        return sshtunnel.SSHTunnelForwarder(
            (self.public_ip(), 22),
            ssh_username=uname,
            ssh_pkey=str(self.ssh_private_key),
            ssh_private_key_password=ssh_key_password,
            local_bind_address=("127.0.0.1", 0),
            remote_bind_address=("127.0.0.1", remote_port),
        )

    def get_ssh_cmd(self, uname="skyplane", ssh_key_password="skyplane"):
        # todo can we include the key password inline?
        return f"ssh -i {self.ssh_private_key} -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no {uname}@{self.public_ip()}"
